@@ -6,7 +6,7 @@ use swc_ecma_ast::*;
 use swc_ecma_transforms_base::{helper, perf::Parallel};
 use swc_ecma_transforms_macros::parallel;
 use swc_ecma_utils::{
-    is_literal, prepend_stmts, private_ident, quote_ident, ExprFactory, StmtLike,
+    is_literal, prepend_stmts, private_ident, quote_ident, undefined, ExprFactory, StmtLike,
 };
 use swc_ecma_visit::{as_folder, noop_visit_mut_type, Fold, VisitMut, VisitMutWith};
 
@@ -64,10 +64,7 @@ impl VisitMut for TemplateLiteral {
                         let should_remove_kind =
                             quasis[0].raw.value.contains('\r') || quasis[0].raw.value.contains('`');
 
-                        let mut s = quasis[0]
-                            .cooked
-                            .clone()
-                            .unwrap_or_else(|| quasis[0].raw.clone());
+                        let mut s = quasis[0].cooked.clone().unwrap();
 
                         // See https://github.com/swc-project/swc/issues/1488
                         //
@@ -319,9 +316,9 @@ impl VisitMut for TemplateLiteral {
                                                 elems: quasis
                                                     .take()
                                                     .into_iter()
-                                                    .map(|elem| {
-                                                        Lit::Str(elem.cooked.unwrap_or(elem.raw))
-                                                            .as_arg()
+                                                    .map(|elem| match elem.cooked {
+                                                        Some(cooked) => Lit::Str(cooked).as_arg(),
+                                                        None => undefined(DUMMY_SP).as_arg(),
                                                     })
                                                     .map(Some)
                                                     .collect(),
